/**
 * The CERN Tape Archive (CTA) project
 * Copyright © 2018 CERN
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "SchemaVersion.hpp"
#include "common/exception/Exception.hpp"
#include "common/make_unique.hpp"

namespace cta { namespace catalogue {

SchemaVersion::SchemaVersion(){}  
  
SchemaVersion::SchemaVersion(const uint64_t schemaVersionMajor, const uint64_t schemaVersionMinor, const cta::optional<uint64_t> schemaVersionMajorNext, const cta::optional<uint64_t> schemaVersionMinorNext, const Status status):m_schemaVersionMajor(schemaVersionMajor), m_schemaVersionMinor(schemaVersionMinor), m_nextSchemaVersionMajor(schemaVersionMajorNext), m_nextSchemaVersionMinor(schemaVersionMinorNext), m_status(status) {
  
}

SchemaVersion::SchemaVersion(const SchemaVersion& orig) {
  if(this != &orig){
    m_schemaVersionMajor = orig.m_schemaVersionMajor;
    m_schemaVersionMinor = orig.m_schemaVersionMinor;
    m_nextSchemaVersionMajor = orig.m_nextSchemaVersionMajor;
    m_nextSchemaVersionMinor = orig.m_nextSchemaVersionMinor;
    m_status = orig.m_status;
  }
}

SchemaVersion & SchemaVersion::operator =(const SchemaVersion& other){
  if(this != &other){
    m_schemaVersionMajor = other.m_schemaVersionMajor;
    m_schemaVersionMinor = other.m_schemaVersionMinor;
    m_nextSchemaVersionMajor = other.m_nextSchemaVersionMajor;
    m_nextSchemaVersionMinor = other.m_nextSchemaVersionMinor;
    m_status = other.m_status;
  }
  return *this;
}

SchemaVersion::~SchemaVersion() {
}

template<>
std::string SchemaVersion::getSchemaVersion() const {
  return std::to_string(m_schemaVersionMajor)+"."+std::to_string(m_schemaVersionMinor);
}

template<>
SchemaVersion::MajorMinor SchemaVersion::getSchemaVersion() const {
  return std::make_pair(m_schemaVersionMajor,m_schemaVersionMinor);
}

template<>
std::string SchemaVersion::getSchemaVersionNext() const {
  std::string schemaVersionNext;
  if(m_nextSchemaVersionMajor && m_nextSchemaVersionMinor){
    schemaVersionNext = std::to_string(m_nextSchemaVersionMajor.value())+"."+std::to_string(m_nextSchemaVersionMinor.value());
  }
  return schemaVersionNext;
}

template<>
SchemaVersion::MajorMinor SchemaVersion::getSchemaVersionNext() const {
  return std::make_pair(m_nextSchemaVersionMajor.value(),m_nextSchemaVersionMinor.value());
}

template<>
std::string SchemaVersion::getStatus() const {
  switch(m_status){
    case COMPLETE:
      return "COMPLETE";
    case UPGRADING:
      return "UPGRADING";
    default:
      throw cta::exception::Exception("In SchemaVersion::getStatus(), wrong status");
  }
}

template<>
SchemaVersion::Status SchemaVersion::getStatus() const {
  return m_status;
}

SchemaVersion::Builder::Builder(){}

SchemaVersion::Builder::Builder(const Builder& other){
  m_schemaVersion = other.m_schemaVersion;
  m_schemaVersionMajorSet = other.m_schemaVersionMajorSet;
  m_schemaVersionMinorSet = other.m_schemaVersionMinorSet;
}

SchemaVersion::Builder & SchemaVersion::Builder::operator=(const Builder& other){
  if(this != &other){
    m_schemaVersion = other.m_schemaVersion;
    m_schemaVersionMajorSet = other.m_schemaVersionMajorSet;
    m_schemaVersionMinorSet = other.m_schemaVersionMinorSet;
  }
  return *this;
}

SchemaVersion::Builder& SchemaVersion::Builder::schemaVersionMajor(const uint64_t pSchemaVersionMajor){
  m_schemaVersion.m_schemaVersionMajor = pSchemaVersionMajor;
  m_schemaVersionMajorSet = true;
  return *this;
}

SchemaVersion::Builder& SchemaVersion::Builder::schemaVersionMinor(const uint64_t pSchemaVersionMinor){
  m_schemaVersion.m_schemaVersionMinor = pSchemaVersionMinor;
  m_schemaVersionMinorSet = true;
  return *this;
}

SchemaVersion::Builder& SchemaVersion::Builder::nextSchemaVersionMajor(const uint64_t pSchemaVersionMajorNext){
  m_schemaVersion.m_nextSchemaVersionMajor = pSchemaVersionMajorNext;
  return *this;
}

SchemaVersion::Builder& SchemaVersion::Builder::nextSchemaVersionMinor(const uint64_t pSchemaVersionMinorNext){
  m_schemaVersion.m_nextSchemaVersionMinor = pSchemaVersionMinorNext;
  return *this;
}

SchemaVersion::Builder& SchemaVersion::Builder::status(const std::string& pstatus){
  try {
    m_schemaVersion.m_status = s_mapStringStatus.at(pstatus);
    return *this;
  } catch(const std::out_of_range& ){
    throw cta::exception::Exception("In SchemaVersion::Builder::status(), wrong status given as parameter: "+pstatus);
  }
}

SchemaVersion::Builder& SchemaVersion::Builder::status(const SchemaVersion::Status & pstatus){
    m_schemaVersion.m_status = pstatus;
    return *this;
}

std::map<std::string,SchemaVersion::Status> SchemaVersion::Builder::s_mapStringStatus {
  {"COMPLETE",Status::COMPLETE},
  {"UPGRADING",Status::UPGRADING}
};

void SchemaVersion::Builder::validate() const { 
  if(!m_schemaVersionMajorSet || !m_schemaVersionMinorSet){
    throw cta::exception::Exception("In SchemaVersion::Builder::validate(), schemaVersionMajor or schemaVersionMinor have not been set.");
  }
  if(m_schemaVersion.m_nextSchemaVersionMajor && m_schemaVersion.m_nextSchemaVersionMinor && m_schemaVersion.m_status == SchemaVersion::Status::COMPLETE){
    throw cta::exception::Exception("In SchemaVersion::Builder::validate(), status is "+m_schemaVersion.getStatus<std::string>()+" but nextSchemaVersionMajor and nextSchemaVersionMinor are defined");
  }
  if(!m_schemaVersion.m_nextSchemaVersionMajor && !m_schemaVersion.m_nextSchemaVersionMinor && m_schemaVersion.m_status == SchemaVersion::Status::UPGRADING){
    throw cta::exception::Exception("In SchemaVersion::Builder::validate(), status is "+m_schemaVersion.getStatus<std::string>()+" but nextSchemaVersionMajor and nextSchemaVersionMinor are NOT defined.");
  }
}

SchemaVersion SchemaVersion::Builder::build() const {
  validate();
  return m_schemaVersion;
}

}}